package LightController;

import java.util.ArrayList;
import java.util.Iterator;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.Condition;

import TronAdapter.*;

public class Dimmer extends TronThread
{
    private enum Loc { UpPassive, UpActive, DnPassive, DnActive }
    Loc location;
    public boolean alive;

    static final int lightOn=1;
    static final int lightOff=0;
    int lightState;

    static final int delay=1001;

    int level;
    int oldLevel;
    int maxLevels;

    int mutantID=0;

    long aTime, nextUpdate;

    ArrayList<LevelListener> listeners;

    Lock lock = null;
    Condition cond = null;

    public Dimmer(long aTime, int levelCount, int mutant)
    {
	super("Dimmer");
	alive = false;
	location = Loc.UpPassive;
	mutantID = mutant;
	level = 0;
	oldLevel = 0;
	lightState = lightOn;
	maxLevels = levelCount;
	this.aTime = aTime;
	listeners = new ArrayList<LevelListener>();
    }

    public void handleTouch()
    {
	synchronized (this) {
	    while (!alive) try {wait();}catch(InterruptedException e){}
	}
	switch (mutantID) {
	case 1:
	    System.out.println("Dimmer: handleTouch!!!");
	    switch (lightState) {
	    case lightOff:
		setLevel(oldLevel);
		lightState = lightOn;
		//break; commented out intentionally for mutant 1
	    case lightOn:
		oldLevel = level;
		setLevel(0);
		lightState = lightOff;
		break;
	    }
	    break;
	default:
	    System.out.println("Dimmer: handleTouch");
	    switch (lightState) {
	    case lightOff:
		setLevel(oldLevel);
		lightState=lightOn;
		break;
	    case lightOn:
		oldLevel = level;
		setLevel(0);
		lightState = lightOff;
		break;
	    }
	    break;
	}
    }

    public void handleStartHold(long startHold)
    {
	synchronized (this) {
	    while (!alive) try {wait();}catch(InterruptedException e){}
	}
	lock.lock();
	nextUpdate = startHold;
	System.out.println("Dimmer: handleStartHold");
	switch (location) {
	case UpPassive:
	    setLevel(oldLevel);
	    lightState = lightOn;
	    location = Loc.UpActive;
	    cond.signalAll();
	    break;
	case DnPassive:
	    //	    setLevel(oldLevel);
	    oldLevel = level;
	    lightState = lightOn;
	    location = Loc.DnActive;
	    cond.signalAll();
	    break;
	default:
	    System.out.println("Dimmer: cannot accept startHold in "+
			       location);
	    //	    interrupt();
	}
	lock.unlock();
    }

    public void handleEndHold()
    {
	synchronized (this) {
	    while (!alive) try {wait();}catch(InterruptedException e){}
	}
	lock.lock();
	System.out.println("Dimmer: handleEndHold");
	switch(location){
	case UpActive:
	    location = Loc.DnPassive;
	    //	    oldLevel = level;
	    cond.signalAll();
	    break;
	case DnActive:
	    location = Loc.UpPassive;
	    //	    oldLevel = level;
	    cond.signalAll();
	    break;
	default:
	    System.out.println("Dimmer: cannot accept endHold in "+location);
	    //	    interrupt();
	}
	lock.unlock();
    }

    public void run()
    {
	lock = new TronLock();
	cond = lock.newCondition();
	lock.lock();
	synchronized (this) { alive = true; notifyAll(); }
	boolean doDelay=true;
	System.out.println("Dimmer running mutant "+mutantID);
	while (alive) {
	    System.out.println("Dimmer state: "+location+" level "+level);
	    switch (location) {
	    case UpPassive:
		try { cond.await(); }
		catch (InterruptedException e) { alive = false; }
		break;
		//startHold happened: state has changed to  dimUpActive
	    case UpActive:
		if (doDelay) try {
		    nextUpdate += delay;
		    //		    cond.await(nextUpdate-getTimeMillis(),
		    cond.await(delay,
			       TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) { alive = false; break; }
		if (location==Loc.UpActive) { // timed out=>increaseLevel
		    if (level<maxLevels) {
			setLevel(level+1);
			doDelay = true;
		    } else if (level==maxLevels) {
			location = Loc.DnActive;
			doDelay = false;
		    }
		}// else endhold has changed our state to DnPassive
		break;
	    case DnPassive:
		try { cond.await(); }
		catch (InterruptedException e) { alive = false; }
		//startHold happened: state has changed to  dimDnActive
		break;
	    case DnActive:
		if (doDelay) try {
		    nextUpdate += delay;
		    //cond.await(nextUpdate-getTimeMillis(),
		    cond.await(delay,
			       TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) { alive = false; break; }
		if (location==Loc.DnActive) { // timed out=>decrease Level
		    if(level>0) {
			setLevel(level-1);
			doDelay = true;
		    } else if (level==0) {
			location = Loc.UpActive;
			if (mutantID==2) doDelay = true;
			else doDelay=false;
		    }
		} // else endhold has changed our state to dimUpPassive
		break;
	    default:
		alive = false;
	    }
	}
	System.out.println("Dimmer interrupted in "+location);
	lock.unlock();
    }

    public synchronized int getLevel()
    {
	while (!alive) try {wait();}catch(InterruptedException e){}
	return level;
    }

    public synchronized int getMaxLevels()
    {
	while (!alive) try {wait();}catch(InterruptedException e){}
	return maxLevels;
    }

    void setLevel(int level)
    {
	this.level = level;
	long time = getTimeMillis()-aTime;
	System.out.println("Dimmer: setLevel to "+level+" at "+time);
	Iterator<LevelListener> it = listeners.iterator();
	while (it.hasNext()) it.next().levelChanged(level);
    }

    public synchronized void addLevelListener(LevelListener listener)
    {
	listeners.add(listener);
    }
}
