package LightController;

import java.awt.Color;
import java.awt.Panel;
import java.awt.Graphics;
import java.awt.Dimension;
import java.awt.Image;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.Condition;

import TronAdapter.*;

public class LevelHistory extends Panel implements LevelListener, Runnable
{
    int max, current, history[]=null, len=0;
    int gridLineCount, gridOffset;
    Image img = null;
    int w, h, phase=0;
    int stepSize = 1;
    private boolean alive = false;

    private synchronized void setAlive(boolean a)
    {
	alive = a; notifyAll();
    }

    public synchronized void waitForReady()
    {
	try { while (!alive) wait(); }
	catch (InterruptedException e) {
	    System.err.println(e); System.exit(1);
	}
    }

    Lock lock = null;
    Condition cond = null;

    public void run()
    {
	try {
	    lock = new TronLock();
	    cond = lock.newCondition();
	    lock.lock();
	    setAlive(true);
	    boolean simtime = (!TronThread.realtime());
	    if (simtime) stepSize = 8;
	    int refresh = stepSize-1;
	    while (true) {
		cond.await(300, TimeUnit.MILLISECONDS);
		if (history!=null) {
		    if (len==w) {
			System.arraycopy(history, 1, history, 0, w-1);
			len--;
		    }
		    history[len] = current; len++;
		}
		if (simtime) {
		    refresh++;
		    if (refresh==stepSize) { drawSelf(); refresh = 0; }
		} else drawSelf();
	    }
	} catch (InterruptedException ex) {
	    System.err.println(ex);
	} finally {
	    lock.unlock();
	}
    }

    public LevelHistory(int maxLevel, int initialLevel)
    {
	super();
	max = maxLevel;
	current = initialLevel;
	gridLineCount = 15;
	w = 0; h = 0;
	TronThread t = new TronThread(this, "LevelHistory");
	t.setPriority(Thread.MIN_PRIORITY);
	t.start();
    }

    public void levelChanged(int level)
    {
	current = level;
    }

    public synchronized void paint(Graphics g){
	if (img != null) {
	    g.drawImage(img, 0, 0, this);
	}
    }

    public synchronized void update(Graphics g){
	if (img != null) {
	    g.drawImage(img, 0, 0, this);
	}
    }

    void drawSelf(){
	Dimension size = getSize();
	if (size.width != w || size.height != h) {
	    w = size.width;
	    h = size.height;
	    int hist[] = new int[w];
	    if (history!=null) {
		if (len>=w) {
		    System.arraycopy(history, len-w, hist, 0, w);
		    len = w;
		} else System.arraycopy(history, 0, hist, 0, len);
	    }
	    history = hist;
	    img = createImage(w, h);
	}
	if (img!=null) {
	    Graphics g = img.getGraphics();
	    phase += stepSize;
	    phase %= ((w-1)/gridLineCount);
	    g.setColor(Color.black);
	    g.fillRect(0, 0, w, h);
	    g.setColor(Color.blue);
	    for (int i=1; i<=gridLineCount; i++) { //vertical grid lines
		int x = i*(w-1)/gridLineCount;
		g.drawLine(x-phase, 0, x-phase, h);
	    }
	    for (int i=1; i<max; i++) { //horizontal grid lines
		int y = i*(h-1)/max;
		g.drawLine(0, y, w, y);
	    }
	    g.drawRect(0, 0, w-1, h-1);
	    g.setColor(Color.yellow);
	    for (int i=0; i<len-1; i++) {
		int v1 = h-1-history[i]*(h-1)/max;
		int v2 = h-1-history[i+1]*(h-1)/max;
		g.drawLine(w-1-len+i, v1, w-len+i, v2);
	    }
	}
	repaint();
    }
}
