package TronAdapter;

import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.TimeUnit;

public class TronLock implements Lock
{
    public final int id;
    protected TronThread owner;
    protected ReentrantLock l = null;

    public TronLock()
    {
	TronThread t = (TronThread)Thread.currentThread();
	id = t.mutexInit();
	if (id < 0) l = new ReentrantLock();
	owner = null;
    }

    /* methods to optimize TronCondition implementation: */
    public TronThread getOwner(){ return owner; }
    public void setOwner(TronThread t) { owner = t; }
    public boolean isLocked(){ return (owner!=null); }

    /* Lock interface methods: */
    public void lock()
    {
	if (l != null) l.lock();
	else {
	    TronThread t = (TronThread)Thread.currentThread();
	    t.mutexLock(this);
	    owner = t;// set reference after gaining mutex
	}
    }

    public void lockInterruptibly() {
	throw new UnsupportedOperationException("not implemented");
    }

    public boolean tryLock() {
	throw new UnsupportedOperationException("not implemented");
    }

    public boolean tryLock(long time, TimeUnit unit) {
	throw new UnsupportedOperationException("not implemented");
    }

    public Condition newCondition()
    {
	if (l != null) return l.newCondition();
	else return new TronCondition(this);
    }

    public void unlock()
    {
	if (l != null) l.unlock();
	else {
	    TronThread t = (TronThread)Thread.currentThread();
	    if (t != owner)
		throw new IllegalMonitorStateException("caller is not owner");
	    owner = null; // clear reference while protected by mutex
	    t.mutexUnlock(this);
	}
    }
}
