package TronAdapter;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import java.util.Vector;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.Condition;


public class TronSender extends TronThread
{
    Lock ioLock = null;
    Condition ioCond = null;
    Vector<int[]> incQueue = new Vector<int[]>();
    Vector<int[]> outQueue = new Vector<int[]>();
    int sentActions = 0;
    int gotAcks = 0;
    boolean abort = true;
    TronAdapter adapter = null;

    protected DataInputStream is = null;
    protected DataOutputStream os = null;

    public TronSender(TronAdapter adapter)
    {
	super("TR.Sender");
	this.adapter = adapter;
	start();
    }

    public synchronized void initialize(DataInputStream i, DataOutputStream o)
    {
	is = i;
	os = o;
	configured = true; notifyAll();
	try {
	    //System.err.println("TR waiting for locks being created");
	    while (ioLock == null) wait();
	    //System.err.println("TR got locks created");
	    if (simtime()) {
		abort = true;
		//System.err.println("TR waiting to finish configuration");
		while (abort) wait();
		//System.err.println("TR done finish configuration");
	    }
	} catch (InterruptedException e) {
	    System.err.println(e);
	    System.exit(-1);
	}
    }

    boolean configured = false;

    public void enqueue(int[] action)
    {
	if (simtime()) {
	    incQueue.addElement(action);// assuming ioLock is held remotely
	    try {// acknowledge input and let the lock to be released:
		os.writeInt(1 | TronReporter.SA_Ack);
		os.flush();
	    } catch (IOException e) {
		System.err.println(e); System.exit(-1);
	    }
	} else {
	    ioLock.lock();
	    incQueue.addElement(action);
	    ioCond.signalAll();
	    ioLock.unlock();
	}
    } /* enqueue */

    public synchronized void gotIOAcks(int amount)
    {
	gotAcks += amount;
	notifyAll();
    } /* gotIOAcks */

    public void transmit(int chan, int[] params) throws IOException
    {
	if (simtime()) {
	    int[] action = new int[params.length+1];
	    action[0] = chan;
	    System.arraycopy(params, 0, action, 1, params.length);
	    outQueue.addElement(action);
	} else {
	    os.writeInt(chan);
	    os.writeShort(params.length);
	    for (int i = 0; i<params.length; ++i)
		os.writeInt(params[i]);
	    os.flush();
	}
    } /* transmit */

    private class Feeder extends TronThread
    {
	public Feeder() { super("TR.Feeder"); }
	public void run()
	{
	    //System.err.println("TR.Feeder before gaining lock");
	    ioLock.lock();
	    //System.err.println("TR.Feeder after gaining lock");
	    try {
		while (true) {
		    //System.err.println("TR.Feeder waiting for input");
		    while (incQueue.isEmpty()) ioCond.await();
		    //System.err.println("TR.Feeder processing input");
		    while (!incQueue.isEmpty()) {
			int[] action = incQueue.remove(0);
			int[] p = new int[action.length-1];
			System.arraycopy(action, 1, p, 0, action.length-1);
			ioLock.unlock();
			adapter.perform(action[0], p);
			ioLock.lock();
		    }
		}
	    } catch (InterruptedException e) {
		System.err.println(e);
		System.exit(-1);
	    }
	    ioLock.unlock();
	} /* run() */
    } /* class Feeder */

    private void disconnect()
    {
	try {
	    configured = false;
	    os.close();
	    is.close();
	} catch (IOException e){}
    }

    public void run()
    {
	synchronized (this) {
	    ioLock = new TronLock();
	    ioCond = ioLock.newCondition();
	    notifyAll();
	}
	new Feeder().start();
	if (realtime()) {
	    synchronized (this) { abort = false; notifyAll(); }
	    return ;
	}
	while (true) {
	    synchronized (this) {
		try { while (!configured) wait(); }
		catch(InterruptedException e) {}
	    }
	    //System.err.println("TR.Sender before gaining lock");
	    ioLock.lock();
	    //System.err.println("TR.Sender after gaining lock");
	    try {
		os.writeByte(TronReporter.SA_IOMutex);
		os.writeInt(((TronLock)ioLock).id);
		os.writeByte(TronReporter.SA_IOCond);
		os.writeInt(((TronCondition)ioCond).id);
		os.flush();
		synchronized (this) { abort = false; notifyAll(); }
		while (!abort) {
		    //System.err.println("TR.Sender waiting for output");
		    while (outQueue.isEmpty() && !abort) ioCond.await();
		    //System.err.println("TR.Sender processing output");
		    if (!outQueue.isEmpty()) {
			sentActions = outQueue.size();
			do {
			    int[] action = outQueue.remove(0);
			    os.writeInt(action[0]);
			    os.writeShort(action.length-1);
			    for (int i = 1; i<action.length; ++i)
				os.writeInt(action[i]);
			} while (!outQueue.isEmpty());
			os.flush();
			synchronized (this) {// freeze the simulated time:
			    while (sentActions > gotAcks)// for all acks
				wait();
			    gotAcks = 0;
			}
			ioCond.signalAll();
		    }
		}
	    } catch (IOException e) {
		System.err.println("TR.Sender: "+e);
		synchronized (this) { abort = false; notifyAll(); }
		configured = false;
		disconnect();
	    } catch (InterruptedException ie) {
		System.err.println("TR.Sender: "+ie);
		configured = false;
		disconnect();
	    } finally {
		ioLock.unlock();
	    } /* try */
	} /* while */
    }/* run() */
} /* class Sender */
